'---------------------------------------------------------------------
'  Copyright (C) Microsoft Corporation.  All rights reserved.
' 
'THIS CODE AND INFORMATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY
'KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
'IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
'PARTICULAR PURPOSE.
'---------------------------------------------------------------------

Option Explicit On
Option Strict On

Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.ComponentModel.Design
Imports System.Drawing
Imports System.Text
Imports System.Windows.Forms
Imports System.Windows.Forms.Design.Behavior

Namespace Extensions

    ''' <summary>
    ''' Provides an abstract base implementation for <see cref="Behavior"/> types
    ''' which have designable rectangular bounds.
    ''' </summary>
    ''' <remarks>The <see cref="BoundsBehavior"/> class manages the details
    ''' of drag/drop transactions. 
    ''' </remarks>
    Friend MustInherit Class BoundsBehavior
        Inherits Behavior

        ''' <summary>
        ''' Caches a reference to the control to which this
        ''' <see cref="Behavior"/> is bound.
        ''' </summary>
        Protected relatedControl As Control

        ''' <summary>
        ''' The <see cref="IServiceProvider"/> which provides access to 
        ''' designer services.
        ''' </summary>
        Private serviceProvider As IServiceProvider

        ''' <summary>
        ''' Backs the <see cref="BehaviorService"/> property.
        ''' </summary>
        Protected behaviorServiceValue As BehaviorService

        ''' <summary>
        ''' The <see cref="IDesignerHost"/> for creating transactions.
        ''' </summary>
        Private designerHostValue As IDesignerHost

        ''' <summary>
        ''' Tracks when a mouse capture behavior has been
        ''' pushed onto the behavior stack.
        ''' </summary>
        Private pushedBehavior As Boolean

        ''' <summary>
        ''' Tracks when a mouse drag operation is active.
        ''' </summary>
        Private isDragging As Boolean

        ''' <summary>
        ''' The currently selected edge of the bounds glyph.
        ''' </summary>
        Protected selectedEdge As EdgeSelection = EdgeSelection.None

        ''' <summary>
        ''' Caches the position of the current drag operation.
        ''' </summary>
        Protected currentPosition As Point

        ''' <summary>
        ''' The transaction representing the current drag operation.
        ''' </summary>
        Private dragTransaction As DesignerTransaction

        ''' <summary>
        ''' The name of the drag transaction.
        ''' </summary>
        Protected transactionName As String

        ''' <summary>
        ''' The bounds-related property attached to this behavior.
        ''' </summary>
        ''' <remarks>The type-specific property name must be assigned by 
        ''' deriving classes.
        ''' </remarks>
        Protected propertyName As String

        ''' <summary>
        ''' Initializes a new instance of the <see cref="BoundsBehavior"/> type.
        ''' </summary>
        ''' <param name="serviceProvider">An <see cref="IServiceProvider"/> which
        ''' provides design-time services. </param>
        Friend Sub New(ByVal serviceProvider As IServiceProvider)
            If serviceProvider IsNot Nothing Then
                Me.serviceProvider = serviceProvider
            Else
                Throw New ArgumentNullException("serviceProvider", "must be assigned")
            End If
        End Sub

        ''' <summary>
        ''' Sets the type-specific bounds-related property by using the 
        ''' given <see cref="PropertyDescriptor"/>.
        ''' </summary>
        ''' <param name="pd">The <see cref="PropertyDescriptor"/> to use for setting
        ''' the bounds-related property.</param>
        ''' <remarks>Deriving classes must provide the implementation for setting
        ''' the bounds-related property specific to the extension type.
        ''' </remarks>
        Protected MustOverride Sub SetBoundsProperty(ByVal pd As PropertyDescriptor)

        ''' <summary>
        ''' Gets a reference to the <see cref="BehaviorService"/>.
        ''' </summary>
        Friend ReadOnly Property BehaviorService() As BehaviorService
            Get
                If Me.behaviorServiceValue Is Nothing Then
                    Me.behaviorServiceValue = CType(Me.serviceProvider.GetService( _
                        GetType(BehaviorService)), BehaviorService)
                End If

                Return Me.behaviorServiceValue
            End Get
        End Property

        Friend ReadOnly Property DesignerHost() As IDesignerHost
            Get
                If Me.designerHostValue Is Nothing Then
                    Me.designerHostValue = CType(Me.serviceProvider.GetService( _
                        GetType(IDesignerHost)), IDesignerHost)
                End If

                Return Me.designerHostValue
            End Get
        End Property

        Public Overrides Function OnMouseDown( _
            ByVal g As Glyph, _
            ByVal button As MouseButtons, _
            ByVal mouseLoc As Point) As Boolean

            If button <> MouseButtons.Left Then
                Return Me.pushedBehavior
            End If

            Dim bg As BoundsGlyph = CType(g, BoundsGlyph)

            Me.InitializeDragOperation(bg, mouseLoc)

            Return False
        End Function

        ''' <summary>
        ''' Initiates a drag operation. 
        ''' </summary>
        ''' <param name="bg">A <see cref="BoundsGlyph"/> which provides state
        ''' for the drag operation.</param>
        ''' <param name="mouseLoc">The location at which the drag operation starts.</param>
        ''' <remarks>The <see cref="InitializeDragOperation"/> method pushes a mouse
        ''' capture behavior onto the behavior stack. It also opens a designer transaction
        ''' for the drag operation.
        ''' </remarks>
        Protected Overridable Sub InitializeDragOperation( _
        ByVal bg As BoundsGlyph, _
        ByVal mouseLoc As Point)
            If bg IsNot Nothing Then
                Me.relatedControl = bg.RelatedControl

                pushedBehavior = True
                Me.BehaviorService.PushCaptureBehavior(Me)

                Me.selectedEdge = bg.SelectedEdge(mouseLoc)

                Me.OpenDragTransaction()
            End If
        End Sub

        Public Overrides Function OnMouseUp(ByVal g As Glyph, ByVal button As MouseButtons) As Boolean
            If Me.isDragging Then
                Me.isDragging = False

                Me.CloseDragTransaction(True)

                Me.OnLoseCapture(g, EventArgs.Empty)
            End If

            Return False
        End Function

        Public Overrides Sub OnLoseCapture(ByVal g As Glyph, ByVal e As EventArgs)
            If Me.pushedBehavior Then
                Me.pushedBehavior = False

                If Me.BehaviorService IsNot Nothing Then
                    If (Me.isDragging) Then
                        Me.isDragging = False

                        Me.CloseDragTransaction(False)
                    End If

                    Me.selectedEdge = EdgeSelection.None

                    Me.BehaviorService.PopBehavior(Me)
                End If
            End If
        End Sub

        ''' <summary>
        ''' Creates a new transaction for a drag operation.
        ''' </summary>
        Protected Overridable Sub OpenDragTransaction()
            Me.dragTransaction = _
                Me.DesignerHost.CreateTransaction(Me.transactionName)
        End Sub

        ''' <summary>
        ''' Closes a drag transaction, with optional commit or cancel.
        ''' </summary>
        ''' <param name="commit"><c>true</c> to commit the transaction; 
        ''' otherwise, <c>false</c>.</param>
        Protected Overridable Sub CloseDragTransaction(ByVal commit As Boolean)
            If Me.dragTransaction IsNot Nothing Then
                Dim t As DesignerTransaction = Me.dragTransaction
                Me.dragTransaction = Nothing

                Using t
                    If commit Then
                        t.Commit()
                    Else
                        t.Cancel()
                    End If
                End Using
            End If
        End Sub

        Public Overrides Function OnMouseMove( _
        ByVal g As Glyph, _
        ByVal button As MouseButtons, _
        ByVal mouseLoc As Point) As Boolean
            If Not (Me.pushedBehavior) Then
                Return False
            End If

            If Not (Me.isDragging) Then
                Me.isDragging = True
            End If

            Me.currentPosition = mouseLoc

            Dim pd As PropertyDescriptor = TypeDescriptor.GetProperties(Me.relatedControl)(Me.propertyName)

            Me.SetBoundsProperty(pd)

            Return True
        End Function

    End Class

End Namespace
